from typing_extensions import Final

import json
import logging
import math
import os
import time
import torch
import torch.distributed as dist
import torch.multiprocessing as mp
from functools import partial
from torch.nn.parallel import DistributedDataParallel as DDP
from typing import Callable, Iterable, Optional, TypeVar

from ptgnn.baseneuralmodel import ModelTrainer
from ptgnn.baseneuralmodel.abstractneuralmodel import AbstractNeuralModel
from ptgnn.baseneuralmodel.modulewithmetrics import ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.amlutils import configure_logging
from ptgnn.baseneuralmodel.utils.data import ShardedLazyDataIterable

TRawDatapoint = TypeVar("TRawDatapoint")
TTensorizedDatapoint = TypeVar("TTensorizedDatapoint")
TNeuralModule = TypeVar("TNeuralModule", bound=ModuleWithMetrics)
ModelType = AbstractNeuralModel[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]

__all__ = ["DistributedModelTrainer"]


class DistributedModelTrainer(ModelTrainer[TRawDatapoint, TTensorizedDatapoint, TNeuralModule]):
    """
    A distributed trainer for `AbstractComponent`s. Used mainly for supervised learning.

    Create a `DistributedModelTrainer` by passing a `AbstractNeuralModel` in the constructor.
    Invoke `train()` to initiate the training loop. The root `TNeuralModule` should return a scalar loss.
    """

    LOGGER: Final = logging.getLogger(__name__)

    def _run_training(
        self,
        distibuted_module: DDP,
        training_tensors,
        epoch,
        device,
        optimizer,
        parallelize,
        scheduler,
        shuffle_input: bool = True,
    ):
        sum_epoch_loss, num_minibatches, num_samples = 0.0, 0, 0
        start_time = time.time()
        distibuted_module.train()

        scaler = torch.cuda.amp.GradScaler(enabled=self._enable_amp)
        try:
            with distibuted_module.join(throw_on_early_termination=True):
                for step_idx, (mb_data, raw_samples) in enumerate(
                    self.model.minibatch_iterator(
                        training_tensors(),
                        device=device,
                        max_minibatch_size=self._minibatch_size,
                        yield_partial_minibatches=False,
                        shuffle_input=shuffle_input,
                        parallelize=parallelize,
                    )
                ):
                    optimizer.zero_grad(
                        set_to_none=True
                    )  # https://pytorch.org/docs/stable/generated/torch.optim.Optimizer.zero_grad.html#torch.optim.Optimizer.zero_grad
                    with torch.cuda.amp.autocast(enabled=self._enable_amp):
                        mb_loss = distibuted_module(**mb_data)
                        if torch.isnan(mb_loss):
                            raise Exception("Loss has a NaN value.")

                    mb_loss = scaler.scale(mb_loss)

                    num_minibatches += 1
                    num_samples += len(raw_samples)
                    sum_epoch_loss += float(mb_loss.detach())

                    mb_loss.backward()

                    if self._clip_gradient_norm is not None:
                        scaler.unscale_(optimizer)
                        torch.nn.utils.clip_grad_norm_(
                            distibuted_module.parameters(), self._clip_gradient_norm
                        )

                    scaler.step(optimizer)
                    scaler.update()
                    if scheduler is not None:
                        scheduler.step(epoch_idx=epoch, epoch_step=step_idx)

        except RuntimeError as re:
            self.LOGGER.info(str(re))
        except Exception as e:
            self.LOGGER.exception("Something went wrong: %s", str(e))
            raise e

        elapsed_time = time.time() - start_time
        self.LOGGER.info(
            "Training complete in %.1fsec [%.2f samples/sec]",
            elapsed_time,
            num_samples / elapsed_time,
        )
        assert (
            num_minibatches > 0
        ), "No training minibatches were created. The minibatch size may be too large or the training dataset size too small."
        self.LOGGER.info("Epoch %i: Train Loss %.2f", epoch + 1, sum_epoch_loss / num_minibatches)
        train_metrics = distibuted_module.module.report_metrics()

        for epoch_hook in self._train_epoch_end_hooks:
            epoch_hook(self.model, distibuted_module.module, epoch, train_metrics)

        if len(train_metrics) > 0:
            self.LOGGER.info("Training Metrics: %s", json.dumps(train_metrics, indent=2))

    def _run_validation(
        self,
        distributed_neural_module: DDP,
        validation_tensors,
        epoch,
        best_target_metric,
        device,
        parallelize,
    ):
        distributed_neural_module.eval()
        sum_epoch_loss, num_minibatches, num_samples = 0.0, 0, 0
        start_time = time.time()
        try:
            with distributed_neural_module.join(), distributed_neural_module.no_sync(), torch.no_grad():
                for mb_data, raw_samples in self.model.minibatch_iterator(
                    validation_tensors(),
                    device=device,
                    max_minibatch_size=self._minibatch_size,
                    yield_partial_minibatches=True,
                    shuffle_input=False,
                    parallelize=parallelize,
                ):
                    with torch.cuda.amp.autocast(enabled=self._enable_amp):
                        mb_loss = distributed_neural_module(**mb_data)
                    num_minibatches += 1
                    num_samples += len(raw_samples)
                    sum_epoch_loss += mb_loss.detach()

                elapsed_time = time.time() - start_time
                assert num_samples > 0, "No validation data was found."

            validation_loss = sum_epoch_loss / num_minibatches

        except RuntimeError as re:
            self.LOGGER.exception("Something went wrong: %s", str(re))

        self.LOGGER.info(
            "Validation complete in %.1fsec [%.2f samples/sec]",
            elapsed_time,
            (num_samples / elapsed_time),
        )
        self.LOGGER.info("Epoch %i: Valid Loss %.2f", epoch + 1, validation_loss)

        validation_metrics = distributed_neural_module.module.report_metrics()
        for epoch_hook in self._validation_epoch_end_hooks:
            epoch_hook(self.model, distributed_neural_module.module, epoch, validation_metrics)
        if len(validation_metrics) > 0:
            self.LOGGER.info("Validation Metrics: %s", json.dumps(validation_metrics, indent=2))

        if self._target_metric is not None:
            target_metric = validation_metrics[self._target_metric]
            target_metric = torch.tensor([target_metric], device=mb_loss.device)
            dist.all_reduce(target_metric)
            target_metric = target_metric.item() / dist.get_world_size()
        else:
            dist.all_reduce(validation_loss)
            validation_loss = validation_loss.item() / dist.get_world_size()
            target_metric = validation_loss

        if self._target_metric_higher_is_better:
            target_metric_improved = target_metric > best_target_metric
        else:
            target_metric_improved = target_metric < best_target_metric

        return target_metric, target_metric_improved, validation_metrics

    def train(
        self,
        training_data: Iterable[TRawDatapoint],
        validation_data: Iterable[TRawDatapoint],
        *,
        show_progress_bar: bool = True,
        validate_on_start: bool = True,
        patience: int = 5,
        initialize_metadata: bool = True,
        parallelize: bool = True,
        use_multiprocessing: bool = True,
        exponential_running_average_factor: float = 0.97,
        device=None,
        store_tensorized_data_in_memory: bool = False,
        shuffle_training_data: bool = True,
    ) -> None:
        raise Exception("Use `distributed_train()` instead of calling `train().")

    def distributed_train(
        self,
        world_size: int,
        training_data: ShardedLazyDataIterable[TRawDatapoint],
        validation_data: ShardedLazyDataIterable[TRawDatapoint],
        *,
        validate_on_start: bool = True,
        patience: int = 5,
        initialize_metadata: bool = True,
        parallelize: bool = True,
        shuffle_training_data: bool = True,
        worker_init: Optional[Callable[["DistributedModelTrainer", int, int], None]] = None,
    ) -> None:
        """
        The training-validation loop for `AbstractNeuralModel`s.

        :param training_data: An iterable that yields the training data for a given shard. Note
            that the data iterator should provide the `set_rank(rank, world_size)` that filters
            the data appropriately.
        :param validation_data: the validation set, as in `training_data`.
        :param validate_on_start: Whether to run a validation loop on start
        :param patience: The number of iterations before early stopping kicks in.
        :param initialize_metadata: If true, initialize the metadata from the training_data. Otherwise,
            assume that the model that is being trained has its metadata already initialized.
        :param parallelize: Bool indicating whether to run in parallel
        :param shuffle_training_data: shuffle the incoming data from `training_data`.
        """
        assert torch.distributed.is_available()

        if initialize_metadata:
            training_data.set_rank(0, 1)  # No sharding currently possible during metadata loading.
            self.load_metadata_and_create_network(training_data, parallelize, False)

        self.LOGGER.info(
            "Model has %s trainable parameters.",
            sum(
                param.numel()
                for param in self.neural_module.parameters(recurse=True)
                if param.requires_grad
            ),
        )

        ### Distributed code starts here
        # Spawn processes here
        mp.spawn(
            self._parallel_training_process,
            args=(
                world_size,
                training_data,
                validation_data,
                parallelize,
                patience,
                shuffle_training_data,
                validate_on_start,
                worker_init,
            ),
            nprocs=world_size,
            join=True,
        )
        self._restore_checkpoint()

    def _parallel_training_process(
        self,
        rank,
        world_size,
        training_data,
        validation_data,
        parallelize,
        patience,
        shuffle_training_data,
        validate_on_start: bool,
        worker_init: Optional[Callable[["DistributedModelTrainer", int, int], None]] = None,
    ):
        assert torch.cuda.is_available(), "No CUDA available. Aborting training."

        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "12355"
        dist.init_process_group(backend="nccl", rank=rank, world_size=world_size)

        if worker_init is not None:
            worker_init(self, rank, world_size)

        self.LOGGER.info(
            f"[{os.getpid()}]: world_size = {dist.get_world_size()}, "
            + f"rank = {dist.get_rank()}, backend={dist.get_backend()}"
        )
        device = dist.get_rank()

        self.neural_module.to(device)
        distributed_neural_module = DDP(self.neural_module, device_ids=[device])

        training_data.set_rank(dist.get_rank(), dist.get_world_size())
        training_tensors = partial(
            self.model.tensorize_dataset,
            training_data,
            parallelize=parallelize,
            use_multiprocessing=True,
        )
        validation_data.set_rank(dist.get_rank(), dist.get_world_size())
        validation_tensors = partial(
            self.model.tensorize_dataset,
            validation_data,
            parallelize=parallelize,
            use_multiprocessing=True,
        )

        optimizer = self._create_optimizer(distributed_neural_module.parameters())

        scheduler = None if self._create_scheduler is None else self._create_scheduler(optimizer)

        for hook in self._training_start_hooks:
            hook(self.model, distributed_neural_module.module, optimizer)

        if self._target_metric_higher_is_better and self._target_metric is not None:
            best_target_metric = -math.inf
        else:
            best_target_metric = math.inf
        if validate_on_start:
            target_metric, improved, _ = self._run_validation(
                distributed_neural_module,
                validation_tensors,
                0,
                best_target_metric,
                device,
                parallelize,
            )
            assert improved
            self.LOGGER.info(f"Initial {self._target_metric or 'Loss'}: {target_metric}")
            best_target_metric = target_metric

        num_epochs_not_improved: int = 0

        for epoch in range(self._max_num_epochs):
            try:
                self._run_training(
                    distributed_neural_module,
                    training_tensors,
                    epoch,
                    device,
                    optimizer,
                    parallelize,
                    scheduler,
                    shuffle_training_data,
                )
            except Exception as e:
                self.LOGGER.exception("Error during training", exc_info=e)
                raise e

            target_metric, target_metric_improved, validation_metrics = self._run_validation(
                distributed_neural_module,
                validation_tensors,
                epoch,
                best_target_metric,
                device,
                parallelize,
            )
            if target_metric_improved:
                num_epochs_not_improved = 0
                if dist.get_rank() == 0:
                    self.LOGGER.info(
                        f"Best performance so far "
                        f"({self._target_metric or 'Loss'}: {target_metric:.3f} from {best_target_metric:.3f}). "
                        "Saving model checkpoint."
                    )
                    self._save_checkpoint()

                best_target_metric = target_metric

                for epoch_hook in self._improved_epoch_end_hooks:
                    epoch_hook(
                        self.model, distributed_neural_module.module, epoch, validation_metrics
                    )
            else:
                num_epochs_not_improved += 1
                if num_epochs_not_improved > patience:
                    self.LOGGER.warning(
                        f"The target metric has not improved for {num_epochs_not_improved} epochs . Stopping."
                    )
                    break

        dist.destroy_process_group()



#!/usr/bin/env python
"""
Usage:
    train.py [options] TRAIN_DATA_PATH VALID_DATA_PATH MODEL_FILENAME

Options:
    --amp                      Enable automatic mixed precision.
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 100]
    --minibatch-size=<size>    The minibatch size. [default: 300]
    --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
    --autorestore              Automatically restore a checkpoint if one exists.
    --sequential-run           Do not parallelize data loading. Makes debugging easier.
    --quiet                    Do not show progress bar.
    --world-size=<int>         The number of GPUs to use (assumes single-node, multi-GPUs). [default: -1]
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""
import random
import torch
from docopt import docopt
from dpu_utils.utils import RichPath, run_and_debug
from functools import partial
from pathlib import Path

from ptgnn.baseneuralmodel.distributedtrainer import DistributedModelTrainer
from ptgnn.baseneuralmodel.utils.amlutils import configure_logging, log_run
from ptgnn.baseneuralmodel.utils.data import ShardedLazyDataIterable
from ptgnn.implementations.typilus.graph2class import Graph2Class
from ptgnn.implementations.typilus.train import create_graph2class_gnn_model


def load_from_folder(path: RichPath, shuffle: bool, rank: int, world_size):
    all_files = [
        p
        for i, p in enumerate(path.get_filtered_files_in_dir("*.jsonl.gz"))
        if i % world_size == rank
    ]

    if shuffle:
        random.shuffle(all_files)
    for file in all_files:
        yield from file.read_as_jsonl()


def create_optimizer(parameters):
    from torch.distributed.optim import ZeroRedundancyOptimizer

    # return torch.optim.Adam(parameters, lr=0.01)
    return ZeroRedundancyOptimizer(parameters, optimizer_class=torch.optim.Adam, lr=0.001)


def log_run_lambda(aml_ctx, fold, model, nn, epoch, metrics):
    """A utility function that can be used with partial(), and can be serialized through multiprocessing."""
    log_run(aml_ctx, fold, model, epoch, metrics)


def worker_init(trainer: DistributedModelTrainer, rank, world_isze):
    try:
        from azureml.core.run import Run

        aml_ctx = Run.get_context()
    except:
        aml_ctx = None

    log_path = configure_logging(aml_ctx, rank=rank)

    trainer.register_train_epoch_end_hook(partial(log_run_lambda, aml_ctx, "train-" + str(rank)))
    trainer.register_validation_epoch_end_hook(
        partial(log_run_lambda, aml_ctx, "valid-" + str(rank))
    )

    def upload_hook(model, nn, epoch, metrics):
        aml_ctx.upload_file(name="model.pkl.gz", path_or_stream=str(trainer._checkpoint_location))
        aml_ctx.upload_file(name="full.log", path_or_stream=log_path)

    if rank == 0 and aml_ctx is not None:
        trainer.register_epoch_improved_end_hook(upload_hook)


def run(arguments):
    azure_info_path = arguments.get("--azure-info", None)
    training_data_path = RichPath.create(arguments["TRAIN_DATA_PATH"], azure_info_path)
    training_data = ShardedLazyDataIterable(
        partial(load_from_folder, training_data_path, shuffle=True)
    )

    validation_data_path = RichPath.create(arguments["VALID_DATA_PATH"], azure_info_path)
    validation_data = ShardedLazyDataIterable(
        partial(load_from_folder, validation_data_path, shuffle=False)
    )

    model_path = Path(arguments["MODEL_FILENAME"])
    assert model_path.name.endswith(".pkl.gz"), "MODEL_FILENAME must have a `.pkl.gz` suffix."

    initialize_metadata = True
    restore_path = arguments.get("--restore-path", None)
    if restore_path:
        initialize_metadata = False
        model, nn = Graph2Class.restore_model(Path(restore_path), device="cpu")
    elif arguments["--autorestore"] and model_path.exists():
        initialize_metadata = False
        model, nn = Graph2Class.restore_model(model_path, device="cpu")
    else:
        nn = None
        model = create_graph2class_gnn_model()

    trainer = DistributedModelTrainer(
        model,
        model_path,
        max_num_epochs=int(arguments["--max-num-epochs"]),
        minibatch_size=int(arguments["--minibatch-size"]),
        optimizer_creator=create_optimizer,
        clip_gradient_norm=1,
        target_validation_metric="Accuracy",
        target_validation_metric_higher_is_better=True,
        enable_amp=arguments["--amp"],
    )
    if nn is not None:
        trainer.neural_module = nn

    world_size = int(arguments["--world-size"])
    if world_size == -1:
        world_size = torch.cuda.device_count()

    trainer.distributed_train(
        world_size,
        training_data,
        validation_data,
        initialize_metadata=initialize_metadata,
        parallelize=not arguments["--sequential-run"],
        validate_on_start=True,
        shuffle_training_data=True,
        patience=10,
        worker_init=worker_init,
    )


if __name__ == "__main__":
    args = docopt(__doc__)
    run_and_debug(lambda: run(args), args.get("--debug", False))



import torch
from torch import nn
from typing import Dict, List, Tuple

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer


class EGCMessagePassingLayer(AbstractMessagePassingLayer):
    """
    An implementation of the Efficient Graph Convolution (EGC-S) layer of

    @misc{tailor2021adaptive,
      title={Adaptive Filters and Aggregator Fusion for Efficient Graph Convolutions},
      author={Shyam A. Tailor and Felix L. Opolka and Pietro Liò and Nicholas D. Lane},
      year={2021},
      eprint={2104.01481},
      note={GNNSys Workshop, MLSys '21 and HAET Workshop, ICLR '21}
    }

    Differences compared to the paper:
    * Allow using a single but different aggregator similar to EGC-M.
    * Different bases per edge-type
    """

    def __init__(
        self,
        input_state_dimension: int,
        output_state_dimension: int,
        num_edge_types: int,
        message_aggregation_function: str,
        num_bases: int = 4,
        num_heads: int = 8,
        dropout_rate: float = 0.0,
    ):
        super().__init__()
        self.__input_state_dim = input_state_dimension
        assert output_state_dimension % num_heads == 0
        self.__aggregation_fn = message_aggregation_function

        self.__num_bases = num_bases
        self.__num_heads = num_heads
        self.__output_state_dim = output_state_dimension

        self.__dropout = nn.Dropout(p=dropout_rate)

        self.__bases = nn.ModuleList(
            [
                nn.Linear(input_state_dimension, num_bases * output_state_dimension, bias=False)
                for _ in range(num_edge_types)
            ]
        )
        self.__weight_coeffs = nn.Linear(input_state_dimension, num_heads * num_bases)

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        assert len(adjacency_lists) == len(self.__bases)
        node_weights = self.__weight_coeffs(node_states).reshape(
            -1, self.__num_heads, self.__num_bases, 1
        )

        all_message_targets, all_messages = [], []
        for edge_type_idx, (adj_list, features, edge_bases_layer) in enumerate(
            zip(adjacency_lists, edge_features, self.__bases)
        ):
            edge_sources_idxs, edge_target_idxs = adj_list
            all_message_targets.append(edge_target_idxs)

            edge_source_states = nn.functional.embedding(edge_sources_idxs, node_states)
            edge_source_states = self.__dropout(edge_source_states)
            all_base_inputs = edge_bases_layer(edge_source_states).reshape(
                -1, self.__num_heads, self.__num_bases, self.__output_state_dim // self.__num_heads
            )
            all_messages.append(all_base_inputs)

        aggregated_messages = self._aggregate_messages(
            messages=torch.cat(all_messages, dim=0),
            message_targets=torch.cat(all_message_targets, dim=0),
            num_nodes=node_states.shape[0],
            aggregation_fn=self.__aggregation_fn,
        )  # [num_nodes, num_heads, num_bases, D]

        return (
            (aggregated_messages * node_weights).sum(axis=-2).reshape(-1, self.__output_state_dim)
        )

    @property
    def input_state_dimension(self) -> int:
        return self.__input_state_dim

    @property
    def output_state_dimension(self) -> int:
        return self.__output_state_dim
